from neuralfoil.gen2_5_architecture.main import (
    get_aero_from_kulfan_parameters,
    get_aero_from_airfoil,
    get_aero_from_coordinates,
    get_aero_from_dat_file,
    bl_x_points,
)

__version__ = "0.2.4"
